local addon = DiminishingReturns
if not addon then return end

-- This allow oUF addons to add support themselves
function addon:DeclareOUF(parent, oUF)
	local db = addon.db:RegisterNamespace(parent, {profile={
		['*'] = {
			enabled = true,
			iconSize = 24,
			direction = 'RIGHT',
			spacing = 2,
			anchorPoint = 'TOPLEFT',
			relPoint = 'BOTTOMLEFT',
			xOffset = 0,
			yOffset = -4,
		}
	}})
	
	local getDatabaseFuncs = {}
	
	-- Frame checking code
	local function CheckFrame(frame)
		local unit = frame:GetAttribute('oUF-guessUnit') or frame.unit
		local refUnit = gsub(unit, "%d+$", "")
		if refUnit == "player" or refUnit == "target" or refUnit == "focus" or refUnit == "arena" or refUnit == "party" or refUnit == "raid" then
			local GetDatabase = getDatabaseFuncs[refUnit]
			if not GetDatabase then
				-- Avoid creating several time the same config
				GetDatabase = function() return db.profile[refUnit], db end
				addon:RegisterFrameConfig(parent..': '..addon.L[refUnit], GetDatabase)
				getDatabaseFuncs[refUnit] = GetDatabase
			end
			return addon:SpawnFrame(frame, frame, GetDatabase)
		end
	end

	-- Check existing frames	
	for i, frame in pairs(oUF.objects) do
		CheckFrame(frame)
	end	
	
	-- Register check for future frames
	oUF:RegisterInitCallback(CheckFrame)
end

-- Scan for declared oUF
for index = 1, GetNumAddOns() do 
	local global = GetAddOnMetadata(index, 'X-oUF')
	if global then
		local parent = GetAddOnInfo(index)
		addon:RegisterAddonSupport(parent, function()
			local oUF = _G[global]
			if oUF then
				addon:DeclareOUF(parent, oUF)
				return 'supported', oUF.version
			end
		end)
	end
end

